/*
 * Saltine Cracker Key Distribution Server 1.5
 * Copyright (C) 1999 aempirei
 * Revised 03feb99
 * Produced by The Ambient Empire
 * see keds.txt for information
 */

#include "saltine.h"

#define MAXCONNECTS 32

#define K_RESTORE_NAME "keds.restore"
#define K_MESSAGE_NAME "keds.message"
#define TIMEDELAY 120
#define shl(a, b) (1 << ((((a) & 1) << 2) | ((b) & 3)))

long kcount(void);

void showstatus(time_t, char);
void showtitle(void);
void showhelp(void);

struct keds_client {
   int  socket;
   char buffer[128];
   struct sockaddr_in info;
};

char bin2char[36] = {
   0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62,
   0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E,
   0x6F, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A
};

int char2bin[128] = {
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,
   33,34,35,-1,-1,-1,-1,-1
};

static char kindex[36][36 >> 1][36 >> 2];

FILE *fs_kstatus;

int main(int argc, char *argv[]) {

   struct sockaddr_in keds;
   struct keds_client kclient[MAXCONNECTS], tmpcli;

   FILE *fs_kfile;

   time_t stime, ltime, seconds;

   fd_set readfds;

   int sa_len;
   int keds_sock;
   int activity;
   int sock_opt = TRUE;
   int checkp = FALSE;
   static char hashlen;

   static char *pwfield;
   static char password[9];
   static char passlen;
   static char array[4];

   char addrbuf[80];
   char *kmsg;

   int update = FALSE;
   
   long i, j, k;
   int  x, y, z;
   
   long *hcheck[2];
   
   showtitle();

   srand(time(0));
   seconds = (time_t)0;
   fs_kstatus = stdout;
   
   bzero(kclient, MAXCONNECTS * sizeof(struct keds_client));

   sa_len = sizeof(struct sockaddr_in);

   fs_kfile = fopen(K_MESSAGE_NAME, "rb");

   i = flen(fs_kfile);
   kmsg = malloc(i + 2);

   fread(kmsg, 1, i, fs_kfile);
   memcpy(&kmsg[i], "$\0", 2);

   for(j = 0; j < i; j++)
      if(kmsg[j] == '$') kmsg[j] = ' ';

   fclose(fs_kfile);

   /* display help if no arguments specified */
   if(argc == 1) { showhelp(); exit(EXIT_FAILURE); }

   if(test_integrity()) exit(EXIT_FAILURE);

   /* ----make sure hash is ok and stuff first----
    * this will usually catch strange accidental command line arguments
    * that could potentially corrupt files and such, for example if someone
    * specifies '*' as an option.
    */
   switch (strlen(argv[1])) {
      case 1:         break; /* option: restore previous session */
      case C3HASHLEN: break; /* option: use POSIX crypt(3) Hash */
      case NTHASHLEN: break; /* option: use NT MD4 Hash */
      default: /* field doesn't match any known algorithm */
         fprintf(stderr, "ERROR: invalid password field '%s'\n", argv[1]);
         showhelp(); exit(EXIT_FAILURE);
   }

   if(argc > 2) {
      fs_kstatus = fopen(argv[2], "at");
      if(fs_kstatus == NULL) {
	 fprintf(stderr, "ERROR: cannot open %s for output\n", argv[2]);
	 exit(EXIT_FAILURE);
      }
      printf("- status output to \"%s\"\n", argv[2]);
   } else printf("- status output to screen\n");

   setvbuf(fs_kstatus, NULL, _IONBF, 0);

   /* read hash field from command line input */
   switch (strlen(argv[1])) {
      case 1: /* restoring previous session */
         if(argv[1][0] == '!') {
            if((fs_kfile = fopen(K_RESTORE_NAME, "rb")) == NULL) {
               fprintf(stderr, "ERROR: cannot restore previous session\n"
                               "       (error opening '%s')\n", K_RESTORE_NAME);
               fclose(fs_kfile); exit(EXIT_FAILURE);
            }

            if(flen(fs_kfile) == 0) {
               fprintf(stderr, "ERROR: cannot restore previous session\n"
                               "       (no restore data found)\n");
               fclose(fs_kfile);
               remove(K_RESTORE_NAME);
               exit(EXIT_FAILURE);
            }
         
            fread(kindex, KEYSPACES >> 3, 1, fs_kfile); /*load scan stats */
            fread(&seconds, sizeof(time_t), 1, fs_kfile); /* load process time */
            hashlen = fgetc(fs_kfile); /* load hash length */
            pwfield = malloc(hashlen + 1);
            bzero(pwfield, hashlen + 1);
            fread(pwfield, hashlen, 1, fs_kfile); /* restore hash field */
            /* restore scan stats */

            fclose(fs_kfile);

            fprintf(fs_kstatus, "- session restored\n");
            showstatus(seconds, hashlen);

            break;
         }
      case C3HASHLEN: /* using POSIX crypt(3) Hash */
         if(c3atob(&argv[1][2]) != NULL) {
            hashlen = C3HASHLEN;
            pwfield = malloc(C3HASHLEN + 1);
            memcpy(pwfield, argv[1], C3HASHLEN + 1);
            fprintf(fs_kstatus, "- hash algorithm: LibDES Crypt(3)\n");
            break;
         }
      case NTHASHLEN: /* using NT MD4 Hash */
         if(ntatob(argv[1]) != NULL) {
            hashlen = NTHASHLEN;
            pwfield = malloc(NTHASHLEN + 1);
            memcpy(pwfield, argv[1], NTHASHLEN + 1);
            for(i = 0; i < NTHASHLEN; i++) pwfield[i] |= 0x20;
            fprintf(fs_kstatus, "- hash algorithm: MD4 NT Hash\n");
            break;
         }
      default: /* field doesn't match any known algorithm */
         fprintf(stderr, "ERROR: invalid password field '%s'\n", argv[1]);
         showhelp(); exit(EXIT_FAILURE);
   }
   
   fprintf(fs_kstatus, "- password field: %s\n", pwfield);

   if(!(keds_sock = socket(AF_INET, SOCK_STREAM, 0))) {
      fprintf(stderr, "ERROR: cannot create socket\n");
      exit(EXIT_FAILURE);
   } else fprintf(fs_kstatus, "- master socket created\n");
   
   if(setsockopt(keds_sock, SOL_SOCKET, SO_REUSEADDR,
		 (char *)&sock_opt, sizeof(sock_opt))) {
      fprintf(stderr, "ERROR: cannot set socket options\n");
      exit(EXIT_FAILURE);
   }

   keds.sin_family      = AF_INET;
   keds.sin_addr.s_addr = INADDR_ANY;
   keds.sin_port        = htons(KEDS_PORT);

   if(bind(keds_sock, (SA *)&keds, sa_len)) {
      fprintf(stderr, "ERROR: cannot bind socket\n");
      exit(EXIT_FAILURE);
   } else fprintf(fs_kstatus, "- socket bound to port %i\n", KEDS_PORT);

   if (listen(keds_sock, MAXCONNECTS)) {
      fprintf(stderr, "ERROR: cannot listen on port\n");
      exit(EXIT_FAILURE);
   } else fprintf(fs_kstatus, "- listening for connection\n");

   stime = ltime = time(NULL) - seconds;
   
   for(;;) {
      FD_ZERO(&readfds);

      FD_SET(keds_sock, &readfds);
      for (i = 0; i < MAXCONNECTS; i++)
	 if (kclient[i].socket > 0)
	    FD_SET(kclient[i].socket, &readfds);

      activity = select(MAXCONNECTS + 3, &readfds, NULL, NULL, NULL);
  
      if ((activity < 0) && (errno != EINTR)) { /* error with select() */ }

      if (FD_ISSET(keds_sock, &readfds)) {
	 tmpcli.socket = accept(keds_sock, (SA *)&tmpcli.info, &sa_len);

	 if(tmpcli.socket < 0) {
	    fprintf(stderr, "ERROR: cannot accept connection\n");
	    exit(EXIT_FAILURE);
	 } else {
	    for (i = 0; i < MAXCONNECTS; i++) {
	       if (kclient[i].socket == 0) {
                  memcpy(&kclient[i], &tmpcli, sizeof(struct keds_client));
		  i = MAXCONNECTS;
	       }
	    }
	 }
      }

      for (i = 0; i < MAXCONNECTS; i++) {
	 if (FD_ISSET(kclient[i].socket, &readfds)) {
	    if (recv(kclient[i].socket, kclient[i].buffer, 128, MSG_PEEK) < 0) {
	       close(kclient[i].socket);
	       kclient[i].socket = 0;
	       kclient[i].buffer[0] = 0;
	       fprintf(stderr, "ERROR: connection closed (%s)\n",
                       ipv4_showname(AF_INET, &kclient[i].info.sin_addr,
                                 addrbuf, sizeof(addrbuf)));
	    } else {
	       kclient[i].buffer[0] = 0;
	       while(strchr(kclient[i].buffer, '=') == NULL) {
		  k = recv(kclient[i].socket, kclient[i].buffer, 128, MSG_PEEK);
		  kclient[i].buffer[k] = 0;
	       }
	       k = recv(kclient[i].socket, kclient[i].buffer, 128, 0);
	       
	       fprintf(fs_kstatus, "- (%s) request: ",
                       ipv4_showname(AF_INET, &kclient[i].info.sin_addr,
                                 addrbuf, sizeof(addrbuf)));

	       j = 0;

	       if(kclient[i].buffer[j] == 'm') {
		  fprintf(fs_kstatus, "[msg]");
		  send(kclient[i].socket, kmsg, strlen(kmsg), 0);
		  j++;
	       }

	       if(kclient[i].buffer[j] == 'h') {
		  fprintf(fs_kstatus, "[h:\"%s\"]", pwfield);
		  array[0] = (char)hashlen + 'A';
		  send(kclient[i].socket, array, 1, 0);
		  send(kclient[i].socket, pwfield, strlen(pwfield), 0);
		  j++;
	       }

	       if(kclient[i].buffer[j] == 'p') {
                  update = TRUE;
                  
                  memcpy(array, &kclient[i].buffer[++j], 3);
                  
                  x = char2bin[(int)array[0]];
		  y = char2bin[(int)array[1]];
		  z = char2bin[(int)array[2]];
                  
		  if(!((x == -1) || (y == -1) || (z == -1))) {
		     fprintf(fs_kstatus, "[p:\"%s\"]", array);
                     kindex[x][y >> 1][z >> 2] |= shl(y, z);
		  }
                  j += 3;
               }

	       if(kclient[i].buffer[j] == 's') {
                  k = kcount();
                  
		  fprintf(fs_kstatus, "[s:\"%3.2f%%\"]",
                          100.0 * (float)k / (float)KEYSPACES);
		  array[0] = (char) (k / 1296)       + 'A';
		  array[1] = (char)((k % 1296) / 36) + 'A';
		  array[2] = (char) (k %   36)       + 'A';
 	          send(kclient[i].socket, array, 3, 0);
		  
                  j++;
	       }

	       if(kclient[i].buffer[j] == 'g') {
		  do {
		     x = rand() % 36;
		     y = rand() % 36;
                     z = rand() % 36;
		  } while((kindex[x][y >> 1][z >> 2] & shl(y, z)) == shl(y, z));

		  array[0] = bin2char[x];
		  array[1] = bin2char[y];
		  array[2] = bin2char[z];
		  fprintf(fs_kstatus, "[g:\"%s\"]", array);
		  
		  send(kclient[i].socket, array, 3, 0);
		  j++;
	       }

	       if(kclient[i].buffer[j] == 'f') {
                  update = checkp = TRUE;
                  ltime -= (time_t)TIMEDELAY;
                  
                  passlen = strlen(&kclient[i].buffer[++j]) - 1;
                  memcpy(password, &kclient[i].buffer[j], passlen);
                  password[(int)passlen] = 0x00;
                  fprintf(fs_kstatus, "[f:\"%s\"]", password);
	       }

	       fprintf(fs_kstatus, "\n");

	       send(kclient[i].socket, "=", 1, 0);
	       close(kclient[i].socket);
	       kclient[i].socket = 0;
	    }
	 }
	 
	 if(kcount() == KEYSPACES) {
	    fprintf(fs_kstatus, "- keyspace exausted\n");
	    fclose(fs_kstatus);
	    exit(EXIT_SUCCESS);
	 }
      }

      if((time(NULL) - ltime >= TIMEDELAY) && update) {
         update = FALSE;
         
         ltime = time(NULL);

	 seconds = time(NULL) - stime;

         showstatus(seconds, hashlen);

         /* Here's where we save the current session stats. */
         fs_kfile = fopen(K_RESTORE_NAME, "wb");
         if(fs_kfile == NULL) {
	    fprintf(stderr, "ERROR: cannot open %s for output\n", argv[2]);
	    exit(EXIT_FAILURE);
         }

         fwrite(kindex, KEYSPACES >> 3, 1, fs_kfile); /* save scan stats */
         fwrite(&seconds, sizeof(time_t), 1, fs_kfile); /* save process time */
         fputc(hashlen, fs_kfile); /* save hash length */
         fwrite(pwfield, hashlen, 1, fs_kfile); /* save hash field */
         fclose(fs_kfile);
      }

      if(checkp) {
         checkp = TRUE;

         if(hashlen == C3HASHLEN) {
            hcheck[0] = crypt3(password, pwfield);
  	    hcheck[1] = c3atob(&pwfield[2]);
            checkp = C3HASHBLEN;
         } else {
            hcheck[0] = nthash(password);
            hcheck[1] = ntatob(pwfield);
            checkp = NTHASHBLEN;
         }
	 
         if(!memcmp(hcheck[0], hcheck[1], checkp)) {
            fprintf(fs_kstatus, "- password found: \"%s\"\n", password);
            fprintf(fs_kstatus, "\n[ COMPLETED ]\n");
            fclose(fs_kstatus);
            exit(EXIT_SUCCESS);
         }
      }
   }

   exit(EXIT_SUCCESS);
}

long kcount() {
   int i, j, k;
   long l = 0L;

   for(i = 0; i < 36; i++)
      for(j = 0; j < 36; j++)
         for(k = 0; k < 36; k++)
            if((kindex[i][j >> 1][k >> 2] & shl(j, k)) == shl(j, k)) l++;

   return l;
}

void showstatus(time_t sec, char htype) {
   long i, j;

   switch(htype) {
      case C3HASHLEN: i = 1679616L; break;
      case NTHASHLEN: i = 60466176L; break;
      default:
         fprintf(stderr, "ERROR: you shouldn't get this error\n");
         exit(EXIT_FAILURE);
   }

   j = kcount();

   fprintf(fs_kstatus, "----[ STATUS ]------------------------\n");
   fprintf(fs_kstatus, "- total time elapsed --  %lid %li:%02li:%02li\n", (sec / 3600L) / 24L, (sec / 3600L) % 24L, (sec % 3600L) / 60L, sec % 60L);
   fprintf(fs_kstatus, "-   segments scanned --  %li\n", j);
   fprintf(fs_kstatus, "- keyspace completed --  %.2f%%\n", 100.0 * j / KEYSPACES);
   fprintf(fs_kstatus, "-   avg. proformance --  %li c/s\n", i / sec * j);
   fprintf(fs_kstatus, "--------------------------------------\n");

}
void showtitle() {
   fprintf(stderr, "\n    ************************************************************************");
   fprintf(stderr, "\n    * Saltine Cracker Key Distribution Server v1.5  Ambient Empire 03feb99 *");
   fprintf(stderr, "\n    * Copyright (C) 1999 aempirei             see keds.txt for information *");
   fprintf(stderr, "\n    ************************************************************************\n\n");
}

void showhelp() {
   fprintf(stderr, "\nUSAGE:  keds <passwordfield> [outputfile]\n\n");
   fprintf(stderr, "  <passwordfield>             13 character field from a POSIX passwd entry\n");
   fprintf(stderr, "                              or 32 character field from an NT sam entry\n");
   fprintf(stderr, "                              ( specify '!' to restore previous session )\n");
   fprintf(stderr, "  [outputfile]                alternate file to store status and messages\n");
   fprintf(stderr, "                              ( default is to the screen )\n");
}
